<?php                   
/*  Anno - a system for interoperation of Moodle with an Annotea server.
    Anno is a modified version of Mooan.
    Anno  Copyright (C) 2007 Jukka Huhta, Juho Iso-Markku, Jarno Laitinen,
                        Timo Myyrylinen, Roger Sandstrm, Miro Wikgren
    Mooan Copyright (C) 2006  Heikki Aitakangas, Mika Holmstrm,
                        Kirsi Jokisalo, Anni Kotilainen, Esko Lehtonen,
                        Paula Urpilainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
/**
 * Library of functions for interfacing with the Annotea server.
 *
 * @author Heikki Aitakangas, Jukka Huhta
 * @license http://www.gnu.org/copyleft/gpl.html GNU General Public License
 *
 * Created by mooan-group August 2006 for the Mooan-software project and
 * modified by anno3-group May 2007 for the Anno-software project
 * at the Department of Computer science, University of Helsinki.
 *
 */


/// Classes

/**
 * AnnoMessageData is a data structure that contains all data from one
 * message and it's replies.
 */
class AnnoMessageData {
    /**
     * String, ID of this message.
     */
    var $id;

    /**
     * String, Name of the creator of this message.
     */
    var $creator;

    /**
     * String, last modified data in W3C date format.
     */
    var $date;

    /**
     * String, Content of the message. Will also temporarily hold the URI to fetch the content
     * from before it does get loaded.
     */
    var $content;

    /**
     * XPointer string, the location if this is a comment.
     */
    var $position;

    /**
     * String, category of the comment, if this is a comment.
     */
    var $category;

    /**
     * Array containing AnnoMessageData objects representing replies to this message.
     * Not set if there are no replies.
     */
    var $replies;
}

/// Functions

/**
 * Gets all comments, and replies to those comments, made to the given
 * document.
 *
 * @param  string $documentID URL of the document to fetch attached comments for
 * @return array  An associative array of $commentID -> $comment_data
 */
function anno_get_comments($documentID) {
    // Get the data specified in Annotea protocol section 2.2 return value for all the comments
    // attached to the document
    $comments = anno_get_comment_base_data($documentID);

    // Loop over the comments and get their reply trees
    foreach ($comments as $comment) {
        // Get any replies
        anno_get_reply_tree($comment);
    }

    return $comments;
}

/**
 * Adds a comment to the given document.
 *
 * @param string $documentID URL of the document that the comment is attached to. Should be
 *                           the output of Moodle qualified_me() function.
 * @param string $text       Content of the comment.
 * @param string $creatorID  User who made the comment
 * @param string $place      XPointer pointing at the region of the document that the
 *                           comment is attached to
 * @param string @category   Category of the comment
 * @return mixed ID of the created comment on success, false on failure
 */
function anno_add_comment($documentID, $text, $creatorID, $place, $category) {

    // Check for existing comments in the same place
    $existingComments = anno_get_comments($documentID);
    foreach ($existingComments as $comment) {
        if (($comment->position==$place) && ($comment->category==$category)) {
            // an identical comment alredy exists!
            return false;
        }
    }

    // Generate the timestamp from current server time
    $timestamp = date(DATE_W3C, time());

    // Lowercase and URL encode the docmentID
    $documentID = rawurlencode(utf8_strtolower($documentID));

    // Add a preceding 'http://' to the documentID. Because bakunin's server wants it to be present
    $documentID = 'http://' . $documentID;

    // Construct the whole category string
    $category = ANNO_COMMENT_CATEGORY_NS . '#' . $category;

    // Encode HTML-relevant characters in the text
    $text = htmlspecialchars($text, ENT_COMPAT, 'UTF-8');

    // Get the text's length
    $textlen = utf8_strlen($text);

    // Construct the request string
    $request = <<<EOT
<?xml version="1.0" ?>
<r:RDF xmlns:r="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
       xmlns:a="http://www.w3.org/2000/10/annotation-ns#"
       xmlns:d="http://purl.org/dc/elements/1.1/"
       xmlns:h="http://www.w3.org/1999/xx/http#">
 <r:Description>
  <r:type r:resource="http://www.w3.org/2000/10/annotation-ns#Annotation"/>
  <r:type r:resource="http://www.w3.org/2000/10/annotationType#Comment"/>
  <r:type r:resource="${category}"/>
  <a:annotates r:resource="${documentID}"/>
  <a:context>
     ${place}
  </a:context>
  <d:title>-</d:title>
  <d:creator>${creatorID}</d:creator>
  <a:created>${timestamp}</a:created>
  <d:date>${timestamp}</d:date>
  <a:body>
   <r:Description>
    <h:ContentType>text/html</h:ContentType>
    <h:ContentLength>${textlen}</h:ContentLength>
    <h:Body r:parseType="Literal">${text}</h:Body>
   </r:Description>
  </a:body>
 </r:Description>
</r:RDF>
EOT;

    // Send the request to Annotea
    $response = anno_annotea_request(HTTP_METH_POST, '', array('Content-Type' => 'application/xml'), $request);

    // Determine what we got as response
    // NULL means something failed -> false
    if ($response == NULL) {
        return false;
    }
    // Return code 201 means success
    else if ($response['code'] == 201) {
        // Parse the newly created comment's ID
        $parse = new DOMDocument();
        $parse->loadXML($response['body']);
        return anno_parse_id($parse->getElementsByTagNameNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'Description')
                                   ->item(0)->attributes->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'about')
                                   ->textContent);
    }
    // Something went wrong
    else {
        /// @todo check $response for different failures
        return false;
    }
}

/**
 * Adds a response to the given message.
 *
 * @param string $rootID    ID of the first message in the response tree
 * @param string $text      Content of the response
 * @param string $creatorID User who made the response
 * @param string $parentID  ID of the message being responded to
 * @return mixed ID of the created response on success, NULL on failure
 */
function anno_add_message($rootID, $text, $creatorID, $parentID) {
    // Generate the timestamp from current server time
    $timestamp = date(DATE_W3C, time());

    // Construct the full rootID and parentID
    $rootID   = ANNO_ANNOTEA_ADDRESS . '/' . $rootID;
    $parentID = ANNO_ANNOTEA_ADDRESS . '/' . $parentID;

    // Encode HTML-relevant characters in the text
    $text = htmlspecialchars($text, ENT_COMPAT, 'UTF-8');

    // Get the text's length
    $textlen = utf8_strlen($text);

    // Construct the request string
    $request = <<<EOT
<?xml version="1.0" ?>
<r:RDF xmlns:r="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
       xmlns:a="http://www.w3.org/2000/10/annotation-ns#"
       xmlns:d="http://purl.org/dc/elements/1.1/"
       xmlns:tr="http://www.w3.org/2001/03/thread#"
       xmlns:h="http://www.w3.org/1999/xx/http#"
       xmlns:rt="http://www.w3.org/2001/12/replyType">
 <r:Description>
  <r:type r:resource="http://www.w3.org/2001/03/thread#Reply"/>
  <tr:root r:resource="${rootID}"/>
  <tr:inReplyTo r:resource="${parentID}"/>
  <d:title>-</d:title>
  <d:creator>${creatorID}</d:creator>
  <a:created>${timestamp}</a:created>
  <d:date>${timestamp}</d:date>
  <a:body>
   <r:Description>
    <h:ContentType>text/html</h:ContentType>
    <h:ContentLength>${textlen}</h:ContentLength>
    <h:Body r:parseType="Literal">${text}</h:Body>
   </r:Description>
  </a:body>
 </r:Description>
</r:RDF>
EOT;

    // Send the request to Annotea
    $response = anno_annotea_request(HTTP_METH_POST, '', array('Content-Type' => 'application/xml'), $request);

    // Determine what we got as response
    // NULL means something failed -> false
    if ($response == NULL) {
        return false;
    }
    // Return code 201 means success
    else if ($response['code'] == 201) {
        // Parse the newly created reply's ID
        $parse = new DOMDocument();
        $parse->loadXML($response['body']);
        return anno_parse_id($parse->getElementsByTagNameNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'Description')
                                    ->item(0)->attributes->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'about')
                                    ->textContent);
    }
    // Something went wrong
    else {
        /// @todo check $response for different failures
        return false;
    }
}

/**
 * Removes an entry with the given ID from Annotea. Will only work on entries that have
 * no responses.
 *
 * @param string $messageID ID of the entry to remove
 * @return boolean  true if the removal is successfull, false otherwise
 */
function anno_delete($messageID) {
    // Send the deletion request
    $toreturn = anno_annotea_request(HTTP_METH_DELETE, '/' . $messageID, NULL, NULL);

    // Determine whether the delete worked OK
    // NULL means something failed -> false
    if ($toreturn == NULL) {
        return false;
    }
    // Return code 200 means success
    else if ($toreturn['code'] == 200) {
        return true;
    }
    // Something went wrong
    else {
        return false;
    }
}

/**
 * Checks if the given comment is made more than ANNO_COMMENT_REMOVAL_TIME ago
 *
 * @param string $messageID ID of the entry to check
 * @return boolean  true if the comment is locked, false otherwise
 */
function anno_is_comment_locked($messageID) {
    // Get the contents
    $response = anno_annotea_request(HTTP_METH_GET, '/' . $messageID, array('Accept' => 'application/xml'), '');

    // NULL means something failed -> false
    if ($response == NULL) {
        return true;
    }
    // Return code 200 means success
    else if ($response['code'] != 200) {
        return true;
    }

    // Parse the response into a DOM tree
    $parse = new DOMDocument();
    $parse->loadXML($response['body']);

    // get the date and parse it
    $date = $parse->getElementsByTagNameNS('http://purl.org/dc/elements/1.1/', 'date')->item(0)->nodeValue;
    $stamp = strtotime($date);
    
    // the comparison. is the comment's time stamp older than ANNO_COMMENT_REMOVAL_TIME minutes
    if ($stamp + (60 * ANNO_COMMENT_REMOVAL_TIME) < time()) {
        return true;
    }
    return false;
}


/**
 * Checks if the given comment has replies.
 *
 * @param string $messageID ID of the entry to check
 * @return boolean true if the comment has replies
 */
function anno_has_replies($messageID) {
    // Retrieve all replies to this message
    $replytree = new AnnoMessageData;
    $replytree->id = $messageID;
    anno_get_reply_tree($replytree, false); // No need to get the bodies here

    if (isset($messagedata->replies)) {
        return true;
    }

    return false;
}


/**
 * Removes an entry with the given ID, and all responses to it, from Annotea.
 *
 * @param string  $messageID ID of the root of the entry subtree to remove
 * @return boolean  true if the whole removal is successfull, false otherwise
 */
function anno_delete_rec($messageID) {
    // Retrieve all replies to this message
    $replytree = new AnnoMessageData;
    $replytree->id = $messageID;
    anno_get_reply_tree($replytree, false); // No need to get the bodies here

    // Start the real recursive deletion
    return anno_delete_rec_aux($replytree);
}

/**
 * The actual recursive deletion, separated from anno_delete_rec due to optimisation reasons.
 *
 * @param object $messagedata AnnoMessageData of the message to delete
 * @return boolean  true if the whole removal is successfull, false otherwise
 */
function anno_delete_rec_aux($messagedata) {
    // If this message has replies, first recurse to them
    if(isset($messagedata->replies)) {
        foreach ($messagedata->replies as $replydata) {
            // Abort deletion on failure. At least bakunin's server outright refuses to delete messages that have existing replies
            // This will cascade down the recursive call chain
            if (anno_delete_rec_aux($replydata) == false) {
                return false;
            }
        }
    }

    // All children are gone, should be safe to remove this entry
    // Return value comes directly from anno_delete()
    return anno_delete($messagedata->id);
}

/**
 * Modifies a comment with the given ID. Works by loading the original message from Annotea
 * and altering it's values based on the parameters. If a parameter is NULL, the original
 * value is kept.
 *
 * @param string $messageID  ID of the comment to modify
 * @param string $documentID URL of the comment to which the comment will now be attached to
 * @param string $text       New content of the comment
 * @param string $creatorID  The user who will now be listed as the creator of the comment
 * @param string $place      XPointer pointing at the location where the comment will now
 *                           be located
 * @return boolean  true if the modification is successfull, false otherwise
 */
function anno_modify_comment($messageID, $documentID, $text, $creatorID, $place) {
    // Retrieve the current RDF of the comment from Annotea
    $response = anno_annotea_request(HTTP_METH_GET, '/' . $messageID, array('Accept' => 'application/xml'), '');

    // Response code other than 200 means failure to get data out of Annotea
    if($response == NULL || $response['code'] != 200) {
        return false;
    }

    // Parse the response into a DOM tree
    $parse = new DOMDocument();
    $parse->loadXML($response['body']);

    // Update the date
    $parse->getElementsByTagNameNS('http://purl.org/dc/elements/1.1/', 'date')->item(0)->nodeValue = date(DATE_W3C, time());

    // Replace the resource being annotated
    if($documentID != NULL) {
        // Lowercase and URL encode the docmentID
        $documentID = rawurlencode(utf8_strtolower($documentID));

        // Add a preceding 'http://' to the documentID. Because bakunin's server wants it to be present
        $documentID = 'http://' . $documentID;

        // Replace the DOM attribute content
        $parse->getElementsByTagNameNS('http://www.w3.org/2000/10/annotation-ns#', 'annotates')
              ->item(0)->attributes->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource')
              ->value
               = $documentID;
    }

    // If no new body was specified, load the old body
    if($text == NULL) {
        $text = anno_get_body($messageID);
    }

    // Always put in a new body. Required by Annotea
    // Encode HTML-relevant characters in the text
    $text = htmlspecialchars($text, ENT_COMPAT, 'UTF-8');

    // Get the text's length
    $textlen = utf8_strlen($text);

    // Crop off the attribute telling the URI of the old body
    $abody = $parse->getElementsByTagNameNS('http://www.w3.org/2000/10/annotation-ns#', 'body')->item(0);
    $abody->removeAttributeNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource');

    // Append an r:Description node to the a:body
    $abody->appendChild($description = $parse->createElementNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'r:Description'));

    // Append h:ContentType, h:ContentLength to r:Description and fill their contents
    $description->appendChild($parse->createElementNS('http://www.w3.org/1999/xx/http#', 'h:ContentType'))
                ->appendChild($parse->createTextNode('text/html'));
    $description->appendChild($parse->createElementNS('http://www.w3.org/1999/xx/http#', 'h:ContentLength'))
                ->appendChild($parse->createTextNode($textlen));

    // Append an h:Body to the r:Description
    $description->appendChild($hbody = $parse->createElementNS('http://www.w3.org/1999/xx/http#', 'h:Body'));

    // Set the r:parseType attribute on h:Body and fill it's contents
    $hbody->setAttributeNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'r:parseType', 'Literal');
    $hbody->nodeValue = $text;

    // Replace the creator of the comment
    if($creatorID != NULL) {
        // Replace the DOM node content
        $parse->getElementsByTagNameNS('http://purl.org/dc/elements/1.1/', 'creator')->item(0)->nodeValue
               = $creatorID;
    }

    // Replace the position of the comment
    if($place != NULL) {
        // Replace the DOM node content
        $parse->getElementsByTagNameNS('http://www.w3.org/2000/10/annotation-ns#', 'context')->item(0)->nodeValue
               = $place;
    }

    // Send the updated comment to Annotea
    $save = $parse->saveXML();

    $response = anno_annotea_request(HTTP_METH_PUT, '/' . $messageID, array('Content-Type' => 'application/xml'), $save);

    // Determine whether the modification succeeded. Response code 200 means it did
    if($response == NULL || $response['code'] != 200) {
        return false;
    } else {
        return true;
    }
}

/**
 * Modifies a response with the given ID. The position of the response in the response
 * tree is not alterable via this function. Works by loading the original message from
 * Annotea and altering it's values based on the parameters. If a parameter is NULL, the
 * original value is kept.
 *
 * @param string $messageID ID of the reponse to modify
 * @param string $text      New content of the response
 * @param string $creatorID The user who will now be listed as the creator of the response
 * @return boolean  true if the modification is successfull, false otherwise
 */
function anno_modify_message($messageID, $text, $creatorID) {
    // Retrieve the current RDF of the message from Annotea
    $response = anno_annotea_request(HTTP_METH_GET, '/' . $messageID, array('Accept' => 'application/xml'), '');

    // Response code other than 200 means failure to get data out of Annotea
    if($response == NULL || $response['code'] != 200) {
        return false;
    }

    // Parse the response into a DOM tree
    $parse = new DOMDocument();
    $parse->loadXML($response['body']);

    // Update the date
    $parse->getElementsByTagNameNS('http://purl.org/dc/elements/1.1/', 'date')->item(0)->nodeValue = date(DATE_W3C, time());

    // If no new body was specified, load the old body
    if($text == NULL) {
        $text = anno_get_body($messageID);
    }

    // Always put in a new body. Required by Annotea
    // Encode HTML-relevant characters in the text
    $text = htmlspecialchars($text, ENT_COMPAT, 'UTF-8');

    // Get the text's length
    $textlen = utf8_strlen($text);

    // Crop off the attribute telling the URI of the old body
    $abody = $parse->getElementsByTagNameNS('http://www.w3.org/2000/10/annotation-ns#', 'body')->item(0);
    $abody->removeAttributeNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource');

    // Append an r:Description node to the a:body
    $abody->appendChild($description = $parse->createElementNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'r:Description'));

    // Append h:ContentType, h:ContentLength to r:Description and fill their contents
    $description->appendChild($parse->createElementNS('http://www.w3.org/1999/xx/http#', 'h:ContentType'))
                ->appendChild($parse->createTextNode("text/html"));
    $description->appendChild($parse->createElementNS('http://www.w3.org/1999/xx/http#', 'h:ContentLength'))
                ->appendChild($parse->createTextNode($textlen));

    // Append an h:Body to the r:Description
    $description->appendChild($hbody = $parse->createElementNS('http://www.w3.org/1999/xx/http#', 'h:Body'));

    // Set the r:parseType attribute on h:Body and fill it's contents
    $hbody->setAttributeNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'r:parseType', 'Literal');
    $hbody->nodeValue = $text;

    // Replace the creator of the message
    if($creatorID != NULL) {
        // Replace the DOM node content
        $parse->getElementsByTagNameNS('http://purl.org/dc/elements/1.1/', 'creator')->item(0)->nodeValue
               = $creatorID;
    }

    // Send the updated message to Annotea
    $save = $parse->saveXML();

    $response = anno_annotea_request(HTTP_METH_PUT, '/' . $messageID, array('Content-Type' => 'application/xml'), $save);

    // Determine whether the modification succeeded. Response code 200 means it did
    if($response == NULL || $response['code'] != 200) {
        return false;
    } else {
        return true;
    }
}

/**
 * Sends the given request via HTTP to the Annotea server and hands back the reponse.
 *
 * @param int    $method  HTTP_METH_* constant. Annotea uses DELETE, GET, POST, PUT
 * @param string $query   The query string to send to the Annotea server
 * @param array  $headers An associative array containing HTTP headers to send. Contents
 *                        should be in format 'Header name' => 'Header value'
 * @param string $content Content of the request
 * @return mixed  An associative array containing response headers and content,
 *                as specified in HttpRequest::getResponseData, with the addition
 *                of the response code being stored under 'code'.
 *                NULL if the request fails.
 */
function anno_annotea_request($method, $query, $headers, $content) {
    // Translate from the global HTTP_METH_* constants to HttpRequest class'
    // constants. No idea why it has a separate copy of that data
    switch ($method) {
        case HTTP_METH_DELETE: $method_class = HttpRequest::METH_DELETE; break;
        case HTTP_METH_GET:    $method_class = HttpRequest::METH_GET;    break;
        case HTTP_METH_POST:   $method_class = HttpRequest::METH_POST;   break;
        case HTTP_METH_PUT:    $method_class = HttpRequest::METH_PUT;    break;

        default:
            // Should never be called with any other HTTP_METH_* value.
            die('anno_annotea_request() called with unhandled $method value: ' . $method);
    }

    // Set up the request
    $request = new HttpRequest(ANNO_ANNOTEA_ADDRESS . $query, $method_class,
        array('timeout' => 10, 'connecttimeout' => 5)
        );
    if ($headers != NULL) {
        $request->addHeaders($headers);
    }
    if ($method == HTTP_METH_POST) {
        $request->setRawPostData($content);
    }
    if ($method == HTTP_METH_PUT) {
        $request->setPutData($content);
    }

    // Send the request
    try {
        $request->send();
    } catch (HttpException $ex) {
        trigger_error('anno_annotea_request(): Exception occurred when sending message: ' . "\n" . $ex, E_USER_WARNING);
        return NULL;
    }

    // Return response data
    $toreturn = $request->getResponseData();
    $toreturn['code'] = $request->getResponseCode();

    return $toreturn;
}

/**
 * Gets the ID, body URI, creator and position of all the comments attached to the
 * given document.
 *
 * @param string $documentID URL of the document to get attached comments' data for
 * @return array  An associative array of commentID => AnnoMessageData
 */
function anno_get_comment_base_data($documentID) {
    // Lowercase and URL encode the docmentID
    $documentID = rawurlencode(utf8_strtolower($documentID));

    // Add a preceding 'http://' to the documentID. Because bakunin's server wants it to be present
    $documentID = 'http://' . $documentID;

    // Retrieve the raw rdf data from annotea
    $response = anno_annotea_request(HTTP_METH_GET, '?w3c_annotates=' . $documentID, array('Accept' => 'application/xml'), '');

    // Create return array
    $toreturn = array();

    // Determine whether there were any comments based on the response code. 200 means there were and they were successfully
    // retrieved. Anything else means either no comments to the given document or some error. In those cases, we just
    // leave the return array blank
    if($response != NULL && $response['code'] == 200) {
        // Parse the response and retrieve all of the nodes describing comments from it
        $parse = new DOMDocument();
        $parse->loadXML($response['body']);
        $rawcomments = $parse->getElementsByTagNameNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'Description');

        // Loop over the raw comments and parse each into a AnnoMessageData object
        foreach($rawcomments as $rawcomment) {
            $comment = new AnnoMessageData;

            // Crop off the beginning of the comment ID
            $comment->id = anno_parse_id($rawcomment->attributes->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                                                                                  'about')->textContent);

            // Loop over the child nodes of the Description node to get the rest of the data
            foreach($rawcomment->childNodes as $child) {
                // NOTE: This may be dangerous. The check should probably be for the whole namespace instead of the mutable shorthand
                switch($child->nodeName) {
                    case 'a:body': {
                        $comment->content =
                            anno_get_body(anno_parse_id(
                                $child->attributes
                                      ->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource')
                                      ->textContent
                                      ));
                        break;
                    }
                    case 'r:type': {
                        $resource =
                            $child->attributes
                                  ->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource')
                                  ->textContent;

                        // See if this type resource describes the comment category
                        if (utf8_strpos($resource, ANNO_COMMENT_CATEGORY_NS) !== false) {
                            // It does, so parse out the comment category
                            $comment->category = utf8_substr($resource, utf8_strrpos($resource, '#') + 1);
                        }
                        break;
                    }
                    case 'a:context':  $comment->position = utf8_trim($child->textContent); break;
                    case 'dc:creator': $comment->creator  = utf8_trim($child->textContent); break;
                    case 'dc:date':    $comment->date     = utf8_trim($child->textContent); break;
                }
            }

            // Insert the parsed comment into return array
            $toreturn[$comment->id] = $comment;
        }
    }

    return $toreturn;
}

/**
 * Gets the content of an entry from Annotea.
 *
 * @param string $messageID ID of the entry whose body to fetch
 * @return string  The content of the entry or a blank string on error
 */
function anno_get_body($messageID) {
    // Retrieve the raw rdf data from annotea
    $response = anno_annotea_request(HTTP_METH_GET, '/body/' . $messageID, array('Accept' => 'application/xml'), '');

    // Determine whether we succeeded at retrieval or not
    if($response != NULL && $response['code'] == 200) {
        // Decode the html special chars encoded during insertion and return the message body
        return utf8_trim(htmlspecialchars_decode($response['body']));
    }
    else {
        /// @todo Return proper error description
        return '';
    }
}

/**
 * Retrieves the response subtree attached to the given entry. The subtree will be attached to
 * the parent node's data structure.
 *
 * @param object  $parent    The AnnoMessageData of the entry to get responses for.
 * @param boolean $getbodies Whether to get the response's bodies
 */
function anno_get_reply_tree(&$parent, $getbodies = true) {
    // Retrieve the raw rdf data from annotea
    $response = anno_annotea_request(HTTP_METH_GET, '?w3c_reply_tree=' . ANNO_ANNOTEA_ADDRESS . '/' . $parent->id,
                                      array('Accept' => 'application/xml'), '');

    // Determine whether there were any replies based on the return code. 200 means there were some,
    // anything else means either no replies or an error. We ignore errors
    if($response != NULL && $response['code'] == 200) {
        // Parse the response and retrieve all of the nodes describing replies from it
        $parse = new DOMDocument();
        $parse->loadXML($response['body']);
        $rawreplies = $parse->getElementsByTagNameNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'Description');

        // Create an arrays for temporary holding of the replies and reply-relationships
        $replyarray  = array($parent->id => $parent);
        $parentarray = array();

        // Parse the raw replies
        foreach($rawreplies as $rawreply) {
            $reply = new AnnoMessageData;

            // Crop off the beginning of the comment ID
            $reply->id = anno_parse_id($rawreply->attributes->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#',
                                                                                'about')->textContent);

            // Loop over the child nodes of the Description node to get the rest of the data
            foreach($rawreply->childNodes as $child) {
                // NOTE: This may be dangerous. The check should probably be for the whole namespace instead of the mutable shorthand
                switch($child->nodeName) {
                    case 't:inReplyTo': {
                        $parentid =
                            anno_parse_id(
                                $child->attributes
                                      ->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource')
                                      ->textContent
                                      );
                        break;
                    }
                    case 'a:body': {
                        if($getbodies) {
                            $reply->content =
                                anno_get_body(anno_parse_id(
                                    $child->attributes
                                          ->getNamedItemNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'resource')
                                          ->textContent
                                          ));
                        }
                        break;
                    }
                    case 'dc:creator':  $reply->creator  = utf8_trim($child->textContent); break;
                    case 'dc:date':     $reply->date     = utf8_trim($child->textContent); break;
                }
            }

            // Store the parsed data
            $replyarray[$reply->id] = $reply;
            if(!isset($parentarray[$parentid])) {
                $parentarray[$parentid] = array();
            }
            $parentarray[$parentid][$reply->id] = $reply;
        }

        // Attach the reply lists to their parents
        foreach($parentarray as $parentid => $replies) {
            $replyarray[$parentid]->replies = $replies;
        }
    }
}

/**
 * Given an Annotea entity ID in the format of an URI, crops off everything preceding
 * the actual entity ID. For example for 'http://localhost/annotea/01234' returns '01234'.
 *
 * @param string $idstring The ID to parse
 * @return string  The actual ID portion of the given string
 */
function anno_parse_id($idstring) {
    $matches = array();
    preg_match('/\/([0-9a-zA-Z]*)$/', $idstring, $matches);
    return $matches[1];
}

/**
 * Retrieves the given message's creator and last modification data from Annotea.
 *
 * @param string $messageID ID of the message to retrieve.
 * @return mixed  A AnnoMessageData containing the parsed message on successfull
 *                retrieval from Annotea, NULL otherwise.
 */
function anno_get_single_message_base_data($messageID) {
    // Retrieve the RDF of the message from Annotea
    $response = anno_annotea_request(HTTP_METH_GET, $messageID, array('Accept' => 'application/xml'), '');

    // Response code other than 200 means failure to get data out of Annotea
    if($response == NULL || $response['code'] != 200) {
        return NULL;
    }

    // Parse the response into a DOM tree
    $parse = new DOMDocument();
    $parse->loadXML($response['body']);

    // Create the data structure to be returned
    $toreturn = new AnnoMessageData;
    $toreturn->id = $messageID;

    // Get the data out of the RDF
    $rawmessage = $parse->getElementsByTagNameNS('http://www.w3.org/1999/02/22-rdf-syntax-ns#', 'Description')->item(0);
    foreach($rawmessage->childNodes as $child) {
        /// @todo This may be dangerous. The check should probably be for the whole namespace instead of the mutable shorthand
        switch($child->nodeName) {
            case 'dc:creator':  $toreturn->creator  = utf8_trim($child->textContent); break;
            case 'dc:date':     $toreturn->date     = utf8_trim($child->textContent); break;
        }
    }
}

?>
