<?php                   
/*  Anno - a system for interoperation of Moodle with an Annotea server.
    Anno is a modified version of Mooan.
    Anno  Copyright (C) 2007 Jukka Huhta, Juho Iso-Markku, Jarno Laitinen,
                        Timo Myyrylinen, Roger Sandstrm, Miro Wikgren
    Mooan Copyright (C) 2006  Heikki Aitakangas, Mika Holmstrm,
                        Kirsi Jokisalo, Anni Kotilainen, Esko Lehtonen,
                        Paula Urpilainen

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/
/**
 * Functions for setting up highlighting of elements in the document based on
 * the XPointers in the comment data.
 *
 * @author Heikki Aitakangas, Timo Myyrylinen
 * @license http://www.gnu.org/copyleft/gpl.html GNU General Public License
 * 
 * Created by mooan-group August 2006 for the Mooan-software project and
 * modified by anno3-group May 2007 for the Anno-software project
 * at the Department of Computer science, University of Helsinki.
 * 
 */


/// Functions

/**
 * Sets span elements to highlight all comments in $commentarray whose location
 * can be found.
 *
 * @param object $document     DOMDocument object into which the captured page has been loaded.
 * @param array  $commentarray Array of AnnoMessageData containing all the comments made to the
 *                             page loaded into $document.
 * @param array  $readbyuser   Array of commentID => date, where the keys list the comments that
 *                             have been read by current user
 */
function anno_place($document, $commentarray, $readbyuser) {
    // Construct a DOMXPath object on the DOMDocument given
    $matcharr = array();
    $counter  = 0;

    // Check which of the comments that have been read should be displayed as
    // having been read (no changes since the reading was logged)
    $tomarkreadarray = Array();
    foreach ($readbyuser as $commentID => $lastreaddate) {
        if (isset($commentarray[$commentID])) {
            // Determine the latest date the comment tree starting at the current comment has been modified
            $lastmodified = anno_get_last_modified($commentarray[$commentID]);

            // See if it should be marked as read
            if (strtotime($lastreaddate) > $lastmodified) {
                $tomarkreadarray[$commentID] = true;
            }
        }
    }

    // Loop over the comments. They are the objects directly stored in $commentarray
    foreach ($commentarray as $comment) {
        // Sanity check on the XPointer
        if(!$comment->position) {
            continue;
        }

        // $comment->place should be of format documenturl#xpointer(..)
        // Get the stuff inside the xpointer part
        if (!preg_match('/xpointer (?P<parens> \\( (?P<innermatch> ( (?>[^()]+) | (?P>parens) )* ) \\) ) /x',
                       $comment->position, $matcharr)) {
            // No XPointer found in the string
            trigger_error('anno_place(): Failed to parse xpointer from string "' . $comment->position . "\"\n", E_USER_WARNING);
            continue;
        }
        $xpointerstr = $matcharr['innermatch'];

        // Generate a boolean determining whether this comment should be marked as read
        $tomarkread = isset($tomarkreadarray[$comment->id]) && $tomarkreadarray[$comment->id];

        // Determine what kind of XPointer it is. For now, we only handle things that could be generated
        // by anno_x_pointer_lib.js. Which means this is probably not exactly standards-compliant.
        if (utf8_strpos($xpointerstr, 'range-to') !== false) {
            // Crosses element borders
            // Parse the range-to parameters
            $params = anno_parse_rangeto($xpointerstr);

            // Sanity check, anno_parse_rangeto() returns false if it fails
            if (!$params) {
                trigger_error('anno_place(): Failed to parse range-to parameters from string "' . $xpointerstr . "\"\n", E_USER_WARNING);
                continue;
            }

            // Hand off to a dedicated function
            anno_do_rangeto($document, $tomarkread, $comment, anno_parse_rangeto($xpointerstr), $counter);
        } else if (utf8_strpos($xpointerstr, 'string-range') !== false) {
            // Does not cross element borders
            // Parse the range-to parameters
            $params = anno_parse_stringrange($xpointerstr);

            // Sanity check, anno_parse_rangeto() returns false if it fails
            if (!$params) {
                trigger_error('anno_place(): Failed to parse string-range parameters from string "' . $xpointerstr . "\"\n",
                              E_USER_WARNING);
                continue;
            }
            // Handle the string-range highlighting
            anno_do_stringrange($document, $tomarkread, $comment, $params, $counter);
        } else {
            // Just a single whole element
            // Find the node pointed at by the XPath
            $targetnode = anno_parse_xpath($document, $xpathstr);

            // Make sure we have a match
            if (!$targetnode) {
                trigger_error('anno_place(): Failed to find element pointed at by xpath "' . $xpathstr . "\"\n",
                              E_USER_WARNING);
                continue;
            }

            // Wrap all the target's #text children in spans
            anno_wrap_children_new_span($targetnodes->item(0), $tomarkread, $comment, $counter);
        }
    }
}


/**
 * Handles a range-to type XPointer. Wraps all of the text within the given
 * range with span nodes, respecting element boundaries and cutting text elements
 * to smaller ones at the range edges if necessary.
 *
 * The algorithm and structure of this function is based on the equivalent
 * in Anno-project.
 *
 * @param object  $document   DOMDocument containing the page being manipulated
 * @param boolean $tomarkread If true, the comment being positioned should be marked
 *                            having been read already
 * @param object  $comment    AnnoMessageData containing the comment being highlighted
 * @param array   $params     The result of $comment->position being parsed
 *                            with anno_parse_rangeto().
 * @param int     $counter    The running counter for keeping generated IDs unique
 */
function anno_do_rangeto($document, $tomarkread, $comment, $params, &$counter) {
    // Get the target nodes
    $startnode = anno_parse_xpath($document, $params['start_xpath']);
    $endnode   = anno_parse_xpath($document, $params['end_xpath']);

    // Make sure we could find both nodes
    if (!$startnode || !$endnode) {
        if (!$startnode)
            trigger_error('anno_do_rangeto(): Failed to find range-to start element from xpath "' . $params['start_xpath'] . "\"\n",
                          E_USER_WARNING);
        if (!$endnode)
            trigger_error('anno_do_rangeto(): Failed to find range-to end element from xpath "' . $params['end_xpath'] . "\"\n",
                          E_USER_WARNING);
        return;
    }

    // Sanity check  
    if ($comment == NULL) {
        return;
    }

    // Special case - startnode and endnode are the same
    if ($startnode->isSameNode($endnode)) {
        // Use the string-range handler function for this instead
        $stringrangeparams = array('xpath'  => $params['start_xpath'],
                                   'string' => '',
                                   'offset' => $params['start_offset'],
                                   'length' => $params['end_offset'] - $params['start_offset']
                                   );
        anno_do_stringrange($document, $tomarkread, $comment, $stringrangeparams, $counter);
        return;
    }

    // Find the first common parent element
    $commonparent = anno_get_common_parent($startnode, $endnode);

    // Start seeking for the range in the subtree beginning with the common parent
    $foundstart = $handledstart = false;
    $foundend   = $handledend   = false;
    $totallen   = 0;
    $stack      = array($commonparent);

    // Loop until we are out of nodes. Finishing the higlighting will break the loop
    while (!empty($stack) && !$handledend) {
        $currentnode = array_pop($stack);

        // If we found either the start node or the end node, mark it and reset the offset calculation counter
        if ($currentnode->isSameNode($startnode)) {
            $foundstart = true;
            $totallen = 0;
        } else if ($currentnode->isSameNode($endnode)) {
            $foundend = true;
            $totallen = 0;
        }

        // Check if current node is previously added comment
        if ($currentnode->nodeName == 'span' && substr($currentnode->getAttribute('id'), 0, strlen(ANNO_PREFIX)) == ANNO_PREFIX && $handledstart) {
            anno_do_overlapping_comment($document, $currentnode, $tomarkread, $comment, $totallen, $params, $counter, $foundstart, $foundend);
        } else {
            // We are only interesting in doing stuff to #text nodes
            if ($currentnode->nodeName == '#text') {

                // We've found the start node, as pointed at by the XPointer, but not yet found the text offset
                // and this one looks to be the one in which the text offset is located
                if ($foundstart && !$handledstart && $params['start_offset'] < $totallen + $currentnode->length) {
                    // Does any if the text in this #text node precede the higlighted section?
                    if ($params['start_offset'] > $totallen) {
                        // Get the portion of the text preceding the annotated part
                        $pretext = utf8_substr($currentnode->data, 0, $params['start_offset'] - $totallen);

                        // Create a #text node out of it and insert it into the document
                        $prenode = $document->createTextNode($pretext);
                        $currentnode->parentNode->insertBefore($prenode, $currentnode);
                    }

                    // Get the portion of the node content that is annotated
                    $annotext = utf8_substr($currentnode->data, $params['start_offset'] - $totallen);

                    // Create a #text node out of it and create a span node to hold it
                    $annonode = $document->createTextNode($annotext);
                    $spannode = anno_create_span_node($document, $tomarkread, $comment, $counter);
                    $spannode->appendChild($annonode);

                    // Replace the #text node with it. This is a workaround for a PHP bug where if
                    // this were an insertBefore() instead, if it is the parent's first child a
                    // #text node containing one line break will be inserted before it when saveHTML()
                    // is called. Which results in fucked up offsets
                    $currentnode->parentNode->replaceChild($spannode, $currentnode);

                    // Mark start position as having been handled
                    $handledstart = true;
                }
                // Something between start and end nodes
                else if ($foundstart && $handledstart) {
                    // Same as the start node, but for the end node
                    if ($foundend && !$handledend && $params['end_offset'] <= $totallen + $currentnode->length) {
                        // Get the portion of this node's content that is annotated and move it to a span
                        $annotext = utf8_substr($currentnode->data, 0, $params['end_offset'] - $totallen);
                        $annonode = $document->createTextNode($annotext);
                        $spannode = anno_create_span_node($document, $tomarkread, $comment, $counter);
                        $spannode->appendChild($annonode);
                
                        // Replace the #text node with the span
                        $currentnode->parentNode->replaceChild($spannode, $currentnode);
                
                        // See if there is anything in the #text after the annotated text
                        if ($params['end_offset'] < $totallen + $currentnode->length) {
                            // Get the portion of the text preceding the annotated part
                            $posttext = utf8_substr($currentnode->data, $params['end_offset'] - $totallen);
                    
                            // Create a #text node out of it and insert it into the document after the previously created
                            // span node
                            $postnode = $document->createTextNode($posttext);
                            $spannode->parentNode->insertBefore($postnode, $spannode->nextSibling);
                        }
                    
                        // Mark end position as having been handled
                        $handledend = true;
                    } else {
                        // Create a new span node
                        $spannode = anno_create_span_node($document, $tomarkread, $comment, $counter);

                        // Place it at the current node's location among the current node's parent's children
                        $currentnode->parentNode->replaceChild($spannode, $currentnode);

                        // Make the #text node the span's child
                        $spannode->appendChild($currentnode);
                    }
                }

                // Increment the length counter
                $totallen += $currentnode->length;
            }

            // Push the current nodes to stack for handling
            if ($currentnode->childNodes) {
                for ($i = $currentnode->childNodes->length - 1; $i >= 0; $i--) {
                    array_push($stack, $currentnode->childNodes->item($i));
                }
            }
        }
    }
}


/**
 * Calculates the deepest node that is a parent of both $node1 and $node2.
 *
 * @param object $node1 DOMNode for which to seek closest commont parent with $node2
 * @param object $node2 DOMNode for which to seek closest commont parent with $node1
 * @return The parent node found
 */
function anno_get_common_parent($node1, $node2) {
    // List the node's parents
    $node1parents = array();
    $node2parents = array();

    $temp = $node1->parentNode;
    while($temp) {
        $node1parents[] = $temp;
        $temp = $temp->parentNode;
    }

    $temp = $node2->parentNode;
    while($temp) {
        $node2parents[] = $temp;
        $temp = $temp->parentNode;
    }

    // Reverse the arrays so seek can begin from the document root
    $node1parents = array_reverse($node1parents);
    $node2parents = array_reverse($node2parents);

    // Start the seek
    $commonparent = null;
    for ($i = 0; $i < count($node1parents) && $i < count($node2parents); $i++) {
        // If the nodes are the same, it's a common parent
        if ($node1parents[$i]->isSameNode($node2parents[$i])) {
            $commonparent = $node1parents[$i];
        } else {
            // Otherwise, the node's paths diverge here, so the current common parent is the last common parent they have
            break;
        }
    }

    return $commonparent;
}

/**
 * Parses the sub-parameters from a range-to type XPointer.
 *
 * @param string The string containing the XPointer range-to to parse
 * @return array On failure false, on success an associative array of
 *               xpath  => the xpath parameter to string-range (1st param)
 *               string => the string parameter to string-range (2nd param)
 *               offset => the offset parameter to string-range (3rd param)
 *               length => the length parameter to string-range (4th param)
 */
function anno_parse_rangeto($string) {
    // $string should be of form
    // start-point(string-range(...))/range-to(end-point(string-range(...)))

    // Extract the string-range parts
    $regex = <<<EOT
/
start-point \s* \(
(?P<stringrange1> .+ )
\) \/ range-to \( end-point \(
(?P<stringrange2> .+ )
\) \)
/x
EOT;
    $matcharr = array();

    // Make sure we have a match
    if (!preg_match($regex, $string, $matcharr)) {
        trigger_error('anno_parse_rangeto(): Failed to parse "' . $string . "\"\n",
                      E_USER_WARNING);
        return false;
    }

    // Extract data from the string-ranges
    $stringrange1 = anno_parse_stringrange($matcharr['stringrange1']);
    $stringrange2 = anno_parse_stringrange($matcharr['stringrange2']);

    // Make sure both parts parsed OK
    if (!$stringrange1 || !$stringrange2) {
        trigger_error('anno_parse_rangeto(): Failed to find both string-ranges from "' . $string . "\"\n",
                      E_USER_WARNING);
        return false;
    }

    // Extract the relevant bits for returning
    $toreturn = array('start_xpath'  => $stringrange1['xpath'],
                      'start_offset' => $stringrange1['offset'],
                      'end_xpath'    => $stringrange2['xpath'],
                      'end_offset'   => $stringrange2['offset']
                      );
    return $toreturn;
}

/**
 * Handles a string-range type XPointer. Wraps all of the text within the given
 * range with span nodes, respecting element boundaries and cutting text elements
 * to smaller ones at the range edges if necessary.
 *
 * The algorithm and structure of this function is based on the equivalent
 * in Anno-project.
 *
 * @param object  $document   DOMDocument containing the page being manipulated
 * @param boolean $tomarkread If true, the comment being positioned should be marked
 *                            having been read already
 * @param object  $comment    AnnoMessageData containing the comment being highlighted
 * @param array   $params     The result of $comment->position being parsed
 *                            with anno_parse_stringrange().
 * @param int     $counter    The running counter for keeping generated IDs unique
 */
function anno_do_stringrange($document, $tomarkread, $comment, $params, &$counter) {
    // Get the element pointed at by the xpath portion
    $targetnode = anno_parse_xpath($document, $params['xpath']);

    // Make sure we have a match
    if (!$targetnode) {
        trigger_error('anno_do_stringrange(): Failed to find string-range target element from xpath "' . $params['xpath'] . "\"\n",
                      E_USER_WARNING);
        return;
    }

    // Sanity check
    if ($comment == NULL) {     
        return;
    }

    // Start the seek for the annotated range
    $foundstart = false;
    $foundend   = false;
    $totallen   = 0;
    $stack      = array($targetnode);

    // Loop until we are out of node to examine or we have finished the job
    while (!empty($stack) && !$foundend) {
        $currentnode = array_pop($stack);

        // Check if current node is previously added comment
        if ($currentnode->nodeName == 'span' && substr($currentnode->getAttribute('id'), 0, strlen(ANNO_PREFIX)) == ANNO_PREFIX) {
            anno_do_overlapping_comment($document, $currentnode, $tomarkread, $comment, $totallen, $params, $counter, $foundstart, $foundend);
        } else {
            // All the other nodes and their children
            // We're only interested in #text nodes
            if ($currentnode->nodeName == '#text') {
                // Could it contain some part of the annotated text?
                if ((!$foundstart && ($params['offset'] < $totallen + $currentnode->length)) || ($foundstart && !$foundend)) {
                    // Seeking start
                    if (!$foundstart) {
                        // Mark start as found
                        $foundstart = true;

                        // See if there is anything in the #text node before the annotated text
                        if ($params['offset'] > $totallen) {
                            // Get the portion of the text preceding the annotated part
                            $pretext = utf8_substr($currentnode->data, 0, $params['offset'] - $totallen);

                            // Create a #text node out of it and insert it into the document before the current #text node
                            $prenode = $document->createTextNode($pretext);
                            $currentnode->parentNode->insertBefore($prenode, $currentnode);
                        }

                        // Get the portion of the node content that is annotated
                        $annotext = utf8_substr($currentnode->data, $params['offset'] - $totallen, $params['length']);

                        // Create a #text node out of it and create a span node to hold it
                        $annonode = $document->createTextNode($annotext);
                        $spannode = anno_create_span_node($document, $tomarkread, $comment, $counter);
                        $spannode->appendChild($annonode);

                        // Replace the #text node with it. This is a workaround for a PHP bug where if
                        // this were an insertBefore() instead, if it is the parent's first child a
                        // #text node containing one line break will be inserted before it when saveHTML()
                        // is called. Which results in faulty offsets.
                        $currentnode->parentNode->replaceChild($spannode, $currentnode);

                        // See if there is anything in the #text after the annotated text
                        if ($params['offset'] + $params['length'] < $totallen + $currentnode->length) {
                            // Found the end node
                            $foundend = true;

                            // Get the portion of the text preceding the annotated part
                            $posttext = utf8_substr($currentnode->data, $params['offset'] - $totallen + $params['length']);

                            // Create a #text node out of it and insert it into the document after the previously created
                            // span node
                            $postnode = $document->createTextNode($posttext);
                            $spannode->parentNode->insertBefore($postnode, $spannode->nextSibling);
                        }
                    }

                    // Still seeking end
                    else if (!$foundend) {
                        // Get the portion of this node's content that is annotated and move it to a span
                        $annotext = utf8_substr($currentnode->data, 0, $params['offset'] + $params['length'] - $totallen);
                        $annonode = $document->createTextNode($annotext);
                        $spannode = anno_create_span_node($document, $tomarkread, $comment, $counter);
                        $spannode->appendChild($annonode);

                        // Replace the #text node with it
                        $currentnode->parentNode->replaceChild($spannode, $currentnode);

                        // See if there is anything in the #text after the annotated text
                        if ($params['offset'] + $params['length'] < $totallen + $currentnode->length) {
                            // Found the end node
                            $foundend = true;

                            // Get the portion of the text preceding the annotated part
                            $posttext = utf8_substr($currentnode->data, $params['offset'] - $totallen + $params['length']);

                            // Create a #text node out of it and insert it into the document after the previously created
                            // span node
                            $postnode = $document->createTextNode($posttext);
                            $spannode->parentNode->insertBefore($postnode, $spannode->nextSibling);
                        }
                    }
                }

                // Update total length
                $totallen += $currentnode->length;
            }

            // Push this node's children to the stack
            if ($currentnode->childNodes) {
                for ($i = $currentnode->childNodes->length - 1; $i >= 0; $i--) {
                    array_push($stack, $currentnode->childNodes->item($i));
                }
            }
        }
    }
}

/**
 * Handles overlapping span nodes. This function is called every time when
 * anno_do_stringrange() or anno_do_rangeto() encounters a previously added
 * comment. This function also calls itself recursively if it encounters
 * previously added overlapping comments.
 *
 * @param object  $document   DOMDocument containing the page being manipulated
 * @param object  $spannode   DOMNode, the parent span node    
 * @param boolean $tomarkread If true, the comment being positioned should be marked
 *                            having been read already
 * @param object  $comment    AnnoMessageData containing the comment being highlighted
 * @param boolean $totallen   Pointer for this comments total lenght from the beginning
 *                            of the target node
 * @param array   $params     The result of $comment->position being parsed
 *                            with anno_parse_stringrange().
 * @param int     $counter    The running counter for keeping generated IDs unique
 * @param boolean $foundstart Pointer. Has this comment found it's start.
 * @param boolean $foundend   Pointer. Has this comment found it's end.
 */
function anno_do_overlapping_comment($document, $spannode, $tomarkread, $comment, &$totallen, $params, &$counter, &$foundstart, &$foundend) {
    // Stack all the span node's children
    $childrenstack[] = NULL;
    if ($spannode->childNodes) {
        for ($i = $spannode->childNodes->length - 1; $i >= 0; $i--) {
            array_push($childrenstack, $spannode->childNodes->item($i));
        }
    }

    // Prepare a new comment with added overlapping data
    $overlappingcomment = clone($comment);
    if ($spannode->nodeName == 'span' && substr($spannode->getAttribute('id'), 0, strlen(ANNO_PREFIX)) == ANNO_PREFIX) {
        $parentid = $spannode->getAttribute('id');
        // parse comment ids from span id
        $idarray = split('_', $parentid);
        $parentid = "";
        for ($i=1; $i < count($idarray)-1; $i++) {  // Last field is counter
            $parentid .= "_".$idarray[$i];
        }
        $overlappingcomment->id .= $parentid;

        // Handle class modifications
        $parentclass = $spannode->getAttribute('class');
        $parentiscontent = false;
        $parentisspelling = false;

        // Handle multitype class
        $classarray = split(' ', $parentclass);
        foreach ($classarray as $class) {
            $classparts = split('_', $class);

            // If any of the overlapping comments is unread, mark the whole highlight as unread
            if ($classparts[count($classparts)-1] == 'unread' && $classparts[count($classparts)-2] != 'spelling') {
                $tomarkread = false;
            }

            // If overlapping comments are of different type, mark overlapping area as multitype.
            if ($comment->category != $classparts[2] && !($comment->category == 'content' && $classparts[2] == 'multiple')) {
                $overlappingcomment->category = 'multitype';
            }
        }
    }

    // Loop until we are out of child nodes to examine or we have finished the job
    while (!empty($childrenstack) && !$foundend) {
        $currentnode = array_pop($childrenstack);
        if ($currentnode != NULL){

            // Check if current node is previously added comment
            if ($currentnode->nodeName == 'span' && substr($currentnode->getAttribute('id'), 0, strlen(ANNO_PREFIX)) == ANNO_PREFIX) {
                // Call recursively this function for current overlapping comment
                anno_do_overlapping_comment($document, $currentnode, $tomarkread, $comment, $totallen, $params, $counter, $foundstart, $foundend);
            } else {
                // We're only interested in #text nodes
                if ($currentnode->nodeName == '#text') {
                    // Could it contain some part of the annotated text?
                    if ((!$foundstart && ($params['offset'] < $totallen + $currentnode->length)) ||
                        ($foundstart && !$foundend)) {
                        // Seeking start
                        if (!$foundstart) {
                            // Mark start as found
                            $foundstart = true;

                            // See if there is anything in the #text node before the annotated text
                            if ($params['offset'] > $totallen) {
                                // Get the portion of the text preceding the annotated part
                                $pretext = utf8_substr($currentnode->data, 0, $params['offset'] - $totallen);

                                // Create a #text node out of it and insert it into the document before the current #text node
                                $prenode = $document->createTextNode($pretext);
                                $currentnode->parentNode->insertBefore($prenode, $currentnode);
                            }

                            // Get the portion of the node content that is annotated
                            $annotext = utf8_substr($currentnode->data, $params['offset'] - $totallen, $params['length']);

                            // Create a #text node out of it and create a span node to hold it
                            $annonode = $document->createTextNode($annotext);
                            $spannode = anno_create_span_node($document, $tomarkread, $overlappingcomment, $counter);
                            $spannode->appendChild($annonode);

                            // Replace the #text node with it. This is a workaround for a PHP bug where if
                            // this were an insertBefore() instead, if it is the parent's first child a
                            // #text node containing one line break will be inserted before it when saveHTML()
                            // is called. Which results in faulty offsets.
                            $currentnode->parentNode->replaceChild($spannode, $currentnode);

                            // See if there is anything in the #text after the annotated text
                            if ($params['offset'] + $params['length'] < $totallen + $currentnode->length) {
                                // Found the end node
                                $foundend = true;

                                // Get the portion of the text preceding the annotated part
                                $posttext = utf8_substr($currentnode->data, $params['offset'] - $totallen + $params['length']);

                                // Create a #text node out of it and insert it into the document after the previously created
                                // span node
                                $postnode = $document->createTextNode($posttext);
                                $spannode->parentNode->insertBefore($postnode, $spannode->nextSibling);
                            }
                        }

                        // Still seeking end
                        else if (!$foundend) {
                            // Get the portion of this node's content that is annotated and move it to a span
                            $annotext = utf8_substr($currentnode->data, 0, $params['offset'] + $params['length'] - $totallen);
                            $annonode = $document->createTextNode($annotext);
                            $spannode = anno_create_span_node($document, $tomarkread, $overlappingcomment, $counter);
                            $spannode->appendChild($annonode);

                            // Replace the #text node with it
                            $currentnode->parentNode->replaceChild($spannode, $currentnode);

                            // See if there is anything in the #text after the annotated text
                            if ($params['offset'] + $params['length'] < $totallen + $currentnode->length) {
                                // Found the end node
                                $foundend = true;

                                // Get the portion of the text preceding the annotated part
                                $posttext = utf8_substr($currentnode->data, $params['offset'] - $totallen + $params['length']);

                                // Create a #text node out of it and insert it into the document after the previously created
                                // span node
                                $postnode = $document->createTextNode($posttext);
                                $spannode->parentNode->insertBefore($postnode, $spannode->nextSibling);
                            }
                        }
                    }

                    // Update total length
                    $totallen += $currentnode->length;
                }
            }
        }
    }
}


/**
 * A function to wrap the massive regexp used to get the string-range paramaters
 * out of a string-range part of an XPointer.
 *
 * @param string The string containing the XPointer string-range to parse
 * @return array On failure false, on success an associative array of
 *               xpath  => the xpath parameter to string-range (1st param)
 *               string => the string parameter to string-range (2nd param)
 *               offset => the offset parameter to string-range (3rd param)
 *               length => the length parameter to string-range (4th param)
 */
function anno_parse_stringrange($string) {
    $toreturn = array();
    $regex = <<<EOT
/ string-range \(

(?P<xpath>      # The XPath parameter
[^,]+
)

\s* , \s*       # The first comma

(?P<string>     # The substring parameter
"[^"]*"
)

(?:             # The optional paremeters
\s* , \s*       # The second comma

(?P<offset>     # The offset parameter
\d+
)

(?:
\s* , \s*       # The third comma

(?P<length>     # The length parameter
\d+
)

)?)?            # Closing the optional parameter subpatterns

\s*             # Any trailing whitespace

\) /x
EOT;

    // Make sure we have a match
    if (!preg_match($regex, $string, $toreturn)) {
        return false;
    }

    // Lowercase the xpath
    $toreturn['xpath'] = utf8_strtolower($toreturn['xpath']);

    return $toreturn;
}

/**
 * Loops over all the text elements in the subtree starting from the target and
 * moves them to be children of span nodes.
 *
 * @param object  $target     DOMNode whose children to move
 * @param boolean $tomarkread If true, the comment being positioned should be marked
 *                            having been read already
 * @param object  $comment    AnnoMessageData of the comment we are highlighting
 * @param int     $counter    Running counter to keep generated span IDs unique
 */
function anno_wrap_children_new_span($target, $tomarkread, $comment, &$counter) {

    $stack = array($target);

    // Loop over the subtree starting from $target in preorder
    while(!empty($stack)) {
        $currentnode = array_pop($stack);

        // Place a new span node between any #text nodes and their initial parents
        if ($currentnode->nodeName == '#text') {

            try {
                $newspan = anno_create_span_node($target->ownerDocument, $tomarkread, $comment, $counter);
                $currentnode->parentNode->insertBefore($newspan, $currentnode);
                $newspan->appendChild($currentnode);
                $currentnode->parentNode->removeChild($currentnode);
            } catch (DOMException $ex) {
                trigger_error('anno_wrap_children_new_span(): Exception occurred while moving a text node:' ."\n" . $ex . "\n",
                              E_USER_WARNING);
            }
        }

        // Insert any children of the current node into the stack
        if ($currentnode->childNodes) {
            for ($i = $currentnode->childNodes->length - 1; $i >= 0; $i--) {
                array_push($stack, $currentnode->childNodes->item($i));
            }
        }
    }
}

/**
 * Creates a new span node and sets it's id attribute and class attributes.
 *
 * @param object  $document   DOMDocument object containing the page we are adding highlighting to.
 * @param boolean $tomarkread If true, the comment being positioned should be marked
 *                            having been read already
 * @param object  $comment    AnnoMessageData of the comment we are highlighting now
 * @param int     $counter    Running counter to make sure all the id attributes are unique.
 *                            Is incremented by this function when used.
 *
 * @return object  The newly created span node or null on failure
 */
function anno_create_span_node($document, $tomarkread, $comment, &$counter) {
    try {
        // Create the new node
        $node = $document->createElement('span');

        $counterval = $counter++;
        $cssclass   = $comment->category;

        // Check how many comments are overlapping
        $numberofcomments = (int)((strlen($comment->id)+1)/11);

        /**
         * @todo This assumes that there are only two types of comments and no
         *       overlapping spelling comments. 
         */
        $multitype = false;
        $additionalclasses = "";
        if ($cssclass == 'multitype') {
            $multitype = true;
            $numberofcomments--;
            if ($numberofcomments == 1) {
                $cssclass = "content";
            }
            $additionalclasses = " anno_highlight_spelling";
            if ($tomarkread) {
                $additionalclasses .= "_read";
            } else {
                $additionalclasses .= "_unread";
            }
        }

        // Change class attribute on overlapping comments
        if ($numberofcomments > 1) {
            // Maximum number of multiple comments that have unique css-style
            if ($numberofcomments > ANNO_MAXIMUM_OVERLAPPING_VISUALIZATION) {
                $numberofcomments = ANNO_MAXIMUM_OVERLAPPING_VISUALIZATION;
            }
            $cssclass = "multiple_$numberofcomments";
        }

        // Set it's ID to "anno_commentid_runningcounter"
        $node->setAttribute('id', 'anno_' . $comment->id . '_' . ($counterval));

        // Set it's class to "anno_highlight_commentcategory" + a suffix dependent on whether the current user has read it
        if ($tomarkread) {
            $node->setAttribute('class', 'anno_highlight_' . $cssclass . '_read' . $additionalclasses);
        } else {
            $node->setAttribute('class', 'anno_highlight_' . $cssclass . '_unread' . $additionalclasses);
        }

        // Set the onMouserOver handler
        if ($numberofcomments == 1 && !$multitype) {
            $node->setAttribute('onmouseover', 'anno_on_mouse_over_comment(event, ' . "'" . $comment->id . "'" . ');');
        } else {
            $node->setAttribute('onmouseover', 'anno_on_mouse_over_comment_multiple(event, ' . "'" . $comment->id . "'" . ');');
        }


        return $node;
    } catch (DOMException $ex) {
        trigger_error('anno_create_span_node(): Exception occurred while creating a span node:' ."\n" . $ex . "\n",
                      E_USER_WARNING);
        return null;
    }
}

/**
 * Gets the UNIX timestamp of when a modification was last made to
 * the comment & response subtree starting with the given comment.
 *
 * @param object $message AnnoMessageData of the root comment of the subtree to examine
 * @return int  The UNIX timestamp of the date when a something was either
 *              modified or added to the subtree.
 */
function anno_get_last_modified($message) {
    // Get this message's last modified date
    $lastmodified = strtotime($message->date);

    // Sanity check to make sure the date was parseable
    if ($lastmodified === false) {
        trigger_error('anno_get_last_modified(): strtotime() failed to parse date: ' . $message->date . "\n",
                      E_USER_WARNING);
        $lastmodified = -1;
    }

    // Compare it with it's replies dates
    if (isset($message->replies)) {
        foreach ($message->replies as $reply) {
            $lastmodified = max($lastmodified, anno_get_last_modified($reply));
        }
    }

    return $lastmodified;
}

/**
 * Determines the element pointed at by the given "XPath".
 * Only handles such XPaths as could be generated by the interface side of our code.
 *
 * @param object $document The DOMDocument object containing the page we are working on
 * @param string $xpath    The XPath to evaluate
 * @return mixed  The DOMElement found once $xpath has been fully evaluated, or false
 *                if something failed.
 */
function anno_parse_xpath($document, $xpath) {
    // Split the XPath string to it's component parts
    $elements = explode('/', $xpath);
    $node = $document;

    foreach ($elements as $element) {
        if ($element == '') { // Ignore empty values generated by explode()
            continue;
        }

        // Parse this part of the xpath string. If the parse fails, abort
        $matcharr = array();
        if(!preg_match('/ (?P<name> [^[]+ )( \[ (?P<index> \d+ ) \] )? /x', $element, $matcharr)) {
            trigger_error('anno_parse_xpath(): Failed to parse xpath section "' . $element . "\"\n", E_USER_WARNING);
            return false;
        }

        // Loop over current node's children, looking for the one targeted by the xpath
        $i = 0;
        $parentnode = $node;
        foreach ($parentnode->childNodes as $child) {
            // See if this is a child node we should be concerned of
            if ($child instanceof DOMElement && strtolower($child->nodeName) == strtolower($matcharr['name'])) {
                // If there is no index number, select the first node of the correct name
                // Or if there is an index, see if this is the correct indexth node
                if (!isset($matcharr['index']) || ++$i == $matcharr['index']) {
                    $node = $child;
                    break;
                }
            }
        }

        // Check to make sure we found a matching child. If not, abort
        if ($parentnode === $node &&
            // Horrible special case - If parent is table and we are seeking for tbody, do not consider not
            // finding the tbody a failure. Some browsers (FF, I'm looking at you) like to generate a tbody
            // in their internal DOM tree even if the source html does not specify such.
            !(strtolower($parentnode->nodeName) == 'table' && strtolower($matcharr['name']) == 'tbody')
            ) {
            trigger_error('anno_parse_xpath(): Failed to find matching node for xpath section "' . $element . "\"\n", E_USER_WARNING);
            return false;
        }
    }

    // Check to make sure we aren't just returning the document element
    if ($node === $document) {
        trigger_error('anno_parse_xpath(): xpath handling failed, would be returning the document element for xpath "' . $xpath . "\"\n",
                      E_USER_WARNING);
        return false;
    }

    // We got this far, so the seek was successfull. Return the node found
    return $node;
}

?>
